# ################################################################################
# Title: HATI hwutils
# Project: HATI
#
# File       : hwutils.tcl
#
# Description: Stablishes the connection to the remote device.
# 
# Licence and Copyright:
# Copyright (c) 2022 Safran Trusted 4D (https://www.safran-group.com/companies/safran-trusted-4d)
# 
# Parameters (2):
#   1. IP address:port
#   2. xvc_mode
#       Possible values are:
#           - "xvc_no":     Do not use the XVC server mode
#           - "xvc_yes":    Use the XVC server mode
#
# Return: None
#
# You might use, distribute and modify this code and its resulting
# binary form under the terms of the LICENSE.txt provided within the
# repository or project.
#
# If you have not received a copy of the LICENSE.txt along with
# this file please write to info<DOT>spain<AT>orolia<DOT>com and consider that
# this file can not be copied and/or distributed in any forms.
# ################################################################################

proc connects_to_remote_device {ip_and_port xvc_mode} {

    puts "\nOpening Vivado Hardware Manager..."
    open_hw_manager -quiet

    if {$ip_and_port eq "none"} {
        puts "Connecting to local Hardware Server..."
        connect_hw_server -quiet
        puts "Opening hardware targets..."
        open_hw_target -quiet
        current_hw_device  get_hw_devices -quiet
        puts "Successfully connected to the local Hardware Server"
    } elseif {$xvc_mode eq "xvc_yes"} {
        puts "Connecting to remote XVC Hardware Server with IP and port: $ip_and_port..."
        connect_hw_server -quiet
        puts "Opening hardware targets..."
        open_hw_target -xvc_url $ip_and_port -quiet
        refresh_hw_device [lindex [get_hw_devices] 1] -quiet
        puts "Successfully connected to the remote XVC Hardware Server"
    } else {
        puts "Connecting to remote Hardware Server with IP:PORT $ip_and_port..."
        connect_hw_server -url $ip_and_port -allow_non_jtag -quiet
        puts "Opening hardware targets..."
        open_hw_target -quiet
        puts "Successfully connected to the remote Hardware Server"
    }

    puts "Detecting JTAG probes..."
    refresh_hw_device [lindex get_hw_devices 0] -quiet
    refresh_hw_device -update_hw_probes false [lindex get_hw_devices 0] -quiet

}


# ###############################################################
# Name: open_hardware.
#
# Description:
#   Open the hardware manager in vivado
#   and discover all the axi master
#   peripherals.
#
# Parameters (1):
#   1. IP address:port
#
# Return: None
# ###############################################################
proc open_hardware {ip_and_port} {

    set hw [get_hw_devices]
    current_hw_device  get_hw_devices -quiet

    if { [llength ${hw}] == 0 } {
        puts "No hardware detected. Second try:"
        close_hardware

        if {$ip_and_port eq "none"} {
            puts "Connecting to local Hardware Server..."
            connect_hw_server -quiet
        } else {
            puts "Connecting to remote Hardware Server..."
            connect_hw_server -url $ip_and_port -allow_non_jtag
        }

        puts "Opening hardware targets..."
        open_hw_target -quiet
        current_hw_device  get_hw_devices -quiet

        puts "Detecting JTAG probes..."
        refresh_hw_device  [lindex get_hw_devices 0] -quiet
        refresh_hw_device -update_hw_probes false [lindex get_hw_devices 0] -quiet

        if { [llength ${hw}] == 0 } {
            puts "No hardware detected, check probes or load a valid bitstream"
            puts "Closing.."
            close_hardware
            exit 1
        }
    }

    puts "\n"
    puts "Devices detected: \n   ${hw}"
    puts "\n"

}


# ###############################################################
# Name: close_hardware.
#
# Description:
#   Closes the hardware target and hardware manager
#
# Parameters: None.
#
# Return: None
# ###############################################################
proc close_hardware {} {

    puts "Closing Vivado Hardware Manager..."
    close_hw_manager -quiet
}

# ###############################################################
# Name: read_from
# Description:
#   Read a data througout AXI
#
# Parameters:
#   - addr: The address to read from.
#   - output: Output format. Parses the read value.
#       Possible values are:
#           - "hex": Return an hexadecimal value.
#           - "int": Return a signed integer.
#           - "str": Return an ASCII string.
#           - "lts": Return a list with each read value.
#
# Return: Returns the read value form "addr" with
#         format "output"
# ###############################################################
proc read_from {addr {output "hex"}} {

    # Includes functions for parsing the data
    source ./tcl_libs/typeutils.tcl; list

    # Create a hw transaction
    create_hw_axi_txn rtn0 [get_hw_axis hw_axi_1] -type read -address [format %08x $addr]
    # Read through AXI
    run_hw_axi -quiet [get_hw_axi_txns rtn0]
    # Get the report. The data read-
    set rd_report [report_hw_axi_txn rtn0 -w 32 -quiet]
    # Delete the transaction.
    delete_hw_axi_txn [get_hw_axi_txns rtn0]

    # rd_report format is { addr rddata }
    # Gets the rddata
    set read_value [lindex ${rd_report} 1]

    # Parses the read data:
    switch ${output} {
        "hex" { set rtnval ${read_value} }
        "int" { set rtnval [eval toSigned ${read_value} "hex"] }
        "str" { set rtnval [toString ${read_value}] }
        "lts" { set rtnval [toList ${read_value}] }
        default {
            puts "ERROR: Invalid Output (${output}) format. Exiting..."
            exit 1
        }
    }

    return ${rtnval}
}


# ###############################################################
# Name: read_sequence
# Description:
#   Read a sequence of data througout AXI
#
# Parameters:
#   - addr: The address to read from.
#   - output: Output format. Parses the read value.
#       Possible values are:
#           - "hex": Return an hexadecimal value.
#           - "int": Return a signed integer.
#           - "str": Return an ASCII string.
#           - "lts": Return a list with each read value.
#
#     - sequence: Number of data to read
#
# Return: <list> Returns a list with all the read data.
# ###############################################################
proc read_sequence {addr {output "hex"} {sequence 1}} {

    # Create an empty list
    set readv {}

    # Read all the consecutive addresses and insert the read data into
    # the list.
    for {set offset 0} { ${offset} < ${sequence} } {incr offset} {
        append readv [eval read_from [expr ${addr} + ${offset} * 0x04] ${output}]
    }

    return ${readv}
}
